@@ -1,12 +1,12 @@
 // Implementation of the builtin builtin.
 #include "config.h"  // IWYU pragma: keep
 
-#include <stddef.h>
+#include "builtin.h"
 
 #include <algorithm>
+#include <cstddef>
 #include <string>
 
-#include "builtin.h"
 #include "builtin_builtin.h"
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
@@ -17,17 +17,20 @@
 struct builtin_cmd_opts_t {
     bool print_help = false;
     bool list_names = false;
+    bool query = false;
 };
-static const wchar_t *const short_options = L":hn";
-static const struct woption long_options[] = {
-    {L"help", no_argument, NULL, 'h'}, {L"names", no_argument, NULL, 'n'}, {NULL, 0, NULL, 0}};
+static const wchar_t *const short_options = L":hnq";
+static const struct woption long_options[] = {{L"help", no_argument, nullptr, 'h'},
+                                              {L"names", no_argument, nullptr, 'n'},
+                                              {L"query", no_argument, nullptr, 'q'},
+                                              {nullptr, 0, nullptr, 0}};
 
 static int parse_cmd_opts(builtin_cmd_opts_t &opts, int *optind, int argc, wchar_t **argv,
                           parser_t &parser, io_streams_t &streams) {
     wchar_t *cmd = argv[0];
     int opt;
     wgetopter_t w;
-    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
+    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
         switch (opt) {
             case 'h': {
                 opts.print_help = true;
@@ -37,6 +40,10 @@ static int parse_cmd_opts(builtin_cmd_opts_t &opts, int *optind, int argc, wchar
                 opts.list_names = true;
                 break;
             }
+            case 'q': {
+                opts.query = true;
+                break;
+            }
             case ':': {
                 builtin_missing_argument(parser, streams, cmd, argv[w.woptind - 1]);
                 return STATUS_INVALID_ARGS;
@@ -69,16 +76,34 @@ int builtin_builtin(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     if (retval != STATUS_CMD_OK) return retval;
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
+        builtin_print_help(parser, streams, cmd);
         return STATUS_CMD_OK;
     }
 
+    if (opts.query && opts.list_names) {
+        streams.err.append_format(BUILTIN_ERR_COMBO2, cmd,
+                                  _(L"--query and --names are mutually exclusive"));
+        return STATUS_INVALID_ARGS;
+    }
+
+    if (opts.query) {
+        wcstring_list_t names = builtin_get_names();
+        retval = STATUS_CMD_ERROR;
+        for (int i = optind; i < argc; i++) {
+            if (contains(names, argv[i])) {
+                retval = STATUS_CMD_OK;
+                break;
+            }
+        }
+        return retval;
+    }
+
     if (opts.list_names) {
         wcstring_list_t names = builtin_get_names();
         std::sort(names.begin(), names.end());
 
-        for (size_t i = 0; i < names.size(); i++) {
-            const wchar_t *el = names.at(i).c_str();
+        for (const auto &name : names) {
+            auto el = name.c_str();
 
             streams.out.append(el);
             streams.out.append(L"\n");
