@@ -2,8 +2,8 @@
 #ifndef FISH_BUILTIN_COMMANDLINE_H
 #define FISH_BUILTIN_COMMANDLINE_H
 
-#include <wchar.h>
 #include <cstring>
+#include <cwchar>
 
 class parser_t;
 
