@@ -1,8 +1,9 @@
 // Implementation of the emit builtin.
 #include "config.h"  // IWYU pragma: keep
 
-#include "builtin.h"
 #include "builtin_emit.h"
+
+#include "builtin.h"
 #include "common.h"
 #include "event.h"
 #include "fallback.h"  // IWYU pragma: keep
@@ -20,7 +21,7 @@ int builtin_emit(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     if (retval != STATUS_CMD_OK) return retval;
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
+        builtin_print_help(parser, streams, cmd);
         return STATUS_CMD_OK;
     }
 
@@ -31,6 +32,6 @@ int builtin_emit(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
 
     const wchar_t *eventname = argv[optind];
     wcstring_list_t args(argv + optind + 1, argv + argc);
-    event_fire_generic(eventname, &args);
+    event_fire_generic(parser, eventname, &args);
     return STATUS_CMD_OK;
 }
