@@ -9,5 +9,5 @@ class parser_t;
 struct io_streams_t;
 
 int builtin_function(parser_t &parser, io_streams_t &streams, const wcstring_list_t &c_args,
-                     const parsed_source_ref_t &source, tnode_t<grammar::job_list> body);
+                     const parsed_source_ref_t &source, tnode_t<grammar::block_statement> body);
 #endif
