@@ -1,15 +1,15 @@
 // Functions used for implementing the set builtin.
 #include "config.h"  // IWYU pragma: keep
 
-#include <errno.h>
-#include <stddef.h>
-#include <stdlib.h>
-#include <string.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include <wchar.h>
 
 #include <algorithm>
+#include <cerrno>
+#include <cstddef>
+#include <cstdlib>
+#include <cstring>
+#include <cwchar>
 #include <iterator>
 #include <memory>
 #include <set>
@@ -21,13 +21,14 @@
 #include "env.h"
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
+#include "history.h"
 #include "io.h"
+#include "parser.h"
 #include "proc.h"
+#include "wcstringutil.h"
 #include "wgetopt.h"
 #include "wutil.h"  // IWYU pragma: keep
 
-class parser_t;
-
 struct set_cmd_opts_t {
     bool print_help = false;
     bool show = false;
@@ -58,14 +59,14 @@ enum {
 // we stop scanning for flags when the first non-flag argument is seen.
 static const wchar_t *const short_options = L"+:LSUaeghlnpqux";
 static const struct woption long_options[] = {
-    {L"export", no_argument, NULL, 'x'},    {L"global", no_argument, NULL, 'g'},
-    {L"local", no_argument, NULL, 'l'},     {L"erase", no_argument, NULL, 'e'},
-    {L"names", no_argument, NULL, 'n'},     {L"unexport", no_argument, NULL, 'u'},
-    {L"universal", no_argument, NULL, 'U'}, {L"long", no_argument, NULL, 'L'},
-    {L"query", no_argument, NULL, 'q'},     {L"show", no_argument, NULL, 'S'},
-    {L"append", no_argument, NULL, 'a'},    {L"prepend", no_argument, NULL, 'p'},
-    {L"path", no_argument, NULL, opt_path}, {L"unpath", no_argument, NULL, opt_unpath},
-    {L"help", no_argument, NULL, 'h'},      {NULL, 0, NULL, 0}};
+    {L"export", no_argument, nullptr, 'x'},    {L"global", no_argument, nullptr, 'g'},
+    {L"local", no_argument, nullptr, 'l'},     {L"erase", no_argument, nullptr, 'e'},
+    {L"names", no_argument, nullptr, 'n'},     {L"unexport", no_argument, nullptr, 'u'},
+    {L"universal", no_argument, nullptr, 'U'}, {L"long", no_argument, nullptr, 'L'},
+    {L"query", no_argument, nullptr, 'q'},     {L"show", no_argument, nullptr, 'S'},
+    {L"append", no_argument, nullptr, 'a'},    {L"prepend", no_argument, nullptr, 'p'},
+    {L"path", no_argument, nullptr, opt_path}, {L"unpath", no_argument, nullptr, opt_unpath},
+    {L"help", no_argument, nullptr, 'h'},      {nullptr, 0, nullptr, 0}};
 
 // Hint for invalid path operation with a colon.
 #define BUILTIN_SET_PATH_ERROR _(L"%ls: Warning: $%ls entry \"%ls\" is not valid (%s)\n")
@@ -86,7 +87,7 @@ static int parse_cmd_opts(set_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncs
 
     int opt;
     wgetopter_t w;
-    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
+    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
         switch (opt) {
             case 'a': {
                 opts.append = true;
@@ -176,42 +177,42 @@ static int validate_cmd_opts(const wchar_t *cmd, set_cmd_opts_t &opts,  //!OCLIN
     // Can't query and erase or list.
     if (opts.query && (opts.erase || opts.list)) {
         streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     // We can't both list and erase variables.
     if (opts.erase && opts.list) {
         streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     // Variables can only have one scope.
     if (opts.local + opts.global + opts.universal > 1) {
         streams.err.append_format(BUILTIN_ERR_GLOCAL, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     // Variables can only have one export status.
     if (opts.exportv && opts.unexport) {
         streams.err.append_format(BUILTIN_ERR_EXPUNEXP, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     // Variables can only have one path status.
     if (opts.pathvar && opts.unpathvar) {
         streams.err.append_format(BUILTIN_ERR_EXPUNEXP, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     // Trying to erase and (un)export at the same time doesn't make sense.
     if (opts.erase && (opts.exportv || opts.unexport)) {
         streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -219,13 +220,13 @@ static int validate_cmd_opts(const wchar_t *cmd, set_cmd_opts_t &opts,  //!OCLIN
     if (opts.show &&
         (opts.local || opts.global || opts.erase || opts.list || opts.exportv || opts.universal)) {
         streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     if (argc == 0 && opts.erase) {
         streams.err.append_format(BUILTIN_SET_ERASE_NO_VAR, cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -235,10 +236,10 @@ static int validate_cmd_opts(const wchar_t *cmd, set_cmd_opts_t &opts,  //!OCLIN
 // Check if we are setting a uvar and a global of the same name exists. See
 // https://github.com/fish-shell/fish-shell/issues/806
 static int check_global_scope_exists(const wchar_t *cmd, set_cmd_opts_t &opts, const wchar_t *dest,
-                                     io_streams_t &streams) {
+                                     io_streams_t &streams, const parser_t &parser) {
     if (opts.universal) {
-        auto global_dest = env_get(dest, ENV_GLOBAL);
-        if (global_dest && shell_is_interactive()) {
+        auto global_dest = parser.vars().get(dest, ENV_GLOBAL);
+        if (global_dest && parser.is_interactive()) {
             streams.err.append_format(BUILTIN_SET_UVAR_ERR, cmd, dest);
         }
     }
@@ -250,7 +251,8 @@ static int check_global_scope_exists(const wchar_t *cmd, set_cmd_opts_t &opts, c
 // contain a colon, then complain. Return true if any path element was valid, false if not.
 static bool validate_path_warning_on_colons(const wchar_t *cmd,
                                             const wchar_t *key,  //!OCLINT(npath complexity)
-                                            const wcstring_list_t &list, io_streams_t &streams) {
+                                            const wcstring_list_t &list, io_streams_t &streams,
+                                            const environment_t &vars) {
     // Always allow setting an empty value.
     if (list.empty()) return true;
 
@@ -266,7 +268,7 @@ static bool validate_path_warning_on_colons(const wchar_t *cmd,
     // not the (missing) local value. Also don't bother to complain about relative paths, which
     // don't start with /.
     wcstring_list_t existing_values;
-    const auto existing_variable = env_get(key, ENV_DEFAULT);
+    const auto existing_variable = vars.get(key, ENV_DEFAULT);
     if (!existing_variable.missing_or_empty()) existing_variable->to_list(existing_values);
 
     for (const wcstring &dir : list) {
@@ -275,7 +277,7 @@ static bool validate_path_warning_on_colons(const wchar_t *cmd,
             continue;
         }
 
-        const wchar_t *colon = wcschr(dir.c_str(), L':');
+        const wchar_t *colon = std::wcschr(dir.c_str(), L':');
         bool looks_like_colon_sep = colon && colon[1];
         if (!looks_like_colon_sep && any_success) {
             // Once we have one valid entry, skip the remaining ones unless we might warn.
@@ -295,62 +297,58 @@ static bool validate_path_warning_on_colons(const wchar_t *cmd,
             any_success = true;
         } else if (looks_like_colon_sep) {
             streams.err.append_format(BUILTIN_SET_PATH_ERROR, cmd, key, dir.c_str(),
-                                      strerror(errno));
+                                      std::strerror(errno));
             streams.err.append_format(BUILTIN_SET_PATH_HINT, cmd, key, key,
-                                      wcschr(dir.c_str(), L':') + 1);
+                                      std::wcschr(dir.c_str(), L':') + 1);
         }
     }
     return any_success;
 }
 
-static void handle_env_return(int retval, const wchar_t *cmd, const wchar_t *key, io_streams_t &streams)
-{
+static void handle_env_return(int retval, const wchar_t *cmd, const wchar_t *key,
+                              io_streams_t &streams) {
     switch (retval) {
         case ENV_OK: {
-            retval = STATUS_CMD_OK;
             break;
         }
         case ENV_PERM: {
             streams.err.append_format(_(L"%ls: Tried to change the read-only variable '%ls'\n"),
                                       cmd, key);
-            retval = STATUS_CMD_ERROR;
             break;
         }
         case ENV_SCOPE: {
             streams.err.append_format(
                 _(L"%ls: Tried to modify the special variable '%ls' with the wrong scope\n"), cmd,
                 key);
-            retval = STATUS_CMD_ERROR;
             break;
         }
         case ENV_INVALID: {
             streams.err.append_format(
-                _(L"%ls: Tried to modify the special variable '%ls' to an invalid value\n"), cmd, key);
-            retval = STATUS_CMD_ERROR;
+                _(L"%ls: Tried to modify the special variable '%ls' to an invalid value\n"), cmd,
+                key);
             break;
         }
         case ENV_NOT_FOUND: {
-            streams.err.append_format(
-                _(L"%ls: The variable '%ls' does not exist\n"), cmd, key);
-            retval = STATUS_CMD_ERROR;
+            streams.err.append_format(_(L"%ls: The variable '%ls' does not exist\n"), cmd, key);
             break;
         }
         default: {
-            DIE("unexpected env_set() ret val");
+            DIE("unexpected vars.set() ret val");
             break;
         }
     }
 }
 
-/// Call env_set. If this is a path variable, e.g. PATH, validate the elements. On error, print a
+/// Call vars.set. If this is a path variable, e.g. PATH, validate the elements. On error, print a
 /// description of the problem to stderr.
 static int env_set_reporting_errors(const wchar_t *cmd, const wchar_t *key, int scope,
-                                    wcstring_list_t &list, io_streams_t &streams) {
-    if (is_path_variable(key) && !validate_path_warning_on_colons(cmd, key, list, streams)) {
+                                    const wcstring_list_t &list, io_streams_t &streams,
+                                    env_stack_t &vars, std::vector<event_t> *evts) {
+    if (is_path_variable(key) && !validate_path_warning_on_colons(cmd, key, list, streams, vars)) {
         return STATUS_CMD_ERROR;
     }
 
-    int retval = env_set(key, scope | ENV_USER, list);
+    int retval = vars.set(key, scope | ENV_USER, list, evts);
     handle_env_return(retval, cmd, key, streams);
 
     return retval;
@@ -365,13 +363,14 @@ static int env_set_reporting_errors(const wchar_t *cmd, const wchar_t *key, int
 /// Returns:
 ///   The total number of indexes parsed, or -1 on error. If any indexes were found the `src` string
 ///   is modified to omit the index expression leaving just the var name.
-static int parse_index(std::vector<long> &indexes, wchar_t *src, int scope, io_streams_t &streams) {
-    wchar_t *p = wcschr(src, L'[');
+static int parse_index(std::vector<long> &indexes, wchar_t *src, int scope, io_streams_t &streams,
+                       const environment_t &vars) {
+    wchar_t *p = std::wcschr(src, L'[');
     if (!p) return 0;  // no slices so nothing for us to do
     *p = L'\0';        // split the var name from the indexes/slices
     p++;
 
-    auto var_str = env_get(src, scope);
+    auto var_str = vars.get(src, scope);
     wcstring_list_t var;
     if (var_str) var_str->to_list(var);
 
@@ -384,7 +383,7 @@ static int parse_index(std::vector<long> &indexes, wchar_t *src, int scope, io_s
             streams.err.append_format(_(L"%ls: Invalid index starting at '%ls'\n"), L"set", src);
             return -1;
         }
-        p = (wchar_t *)end;
+        p = const_cast<wchar_t *>(end);
 
         // Convert negative index to a positive index.
         if (l_ind < 0) l_ind = var.size() + l_ind + 1;
@@ -395,7 +394,7 @@ static int parse_index(std::vector<long> &indexes, wchar_t *src, int scope, io_s
             if (errno > 0) {  // ignore errno == -1 meaning the int did not end with a '\0'
                 return -1;
             }
-            p = (wchar_t *)end;
+            p = const_cast<wchar_t *>(end);
 
             // Convert negative index to a positive index.
             if (l_ind2 < 0) l_ind2 = var.size() + l_ind2 + 1;
@@ -425,7 +424,7 @@ static int update_values(wcstring_list_t &list, std::vector<long> &indexes,
         if (ind < 0) {
             return 1;
         }
-        if ((size_t)ind >= list.size()) {
+        if (static_cast<size_t>(ind) >= list.size()) {
             list.resize(ind + 1);
         }
 
@@ -446,7 +445,7 @@ static void erase_values(wcstring_list_t &list, const std::vector<long> &indexes
     decltype(indexes_set)::const_reverse_iterator iter;
     for (iter = indexes_set.rbegin(); iter != indexes_set.rend(); ++iter) {
         long val = *iter;
-        if (val > 0 && (size_t)val <= list.size()) {
+        if (val > 0 && static_cast<size_t>(val) <= list.size()) {
             // One-based indexing!
             list.erase(list.begin() + val - 1);
         }
@@ -466,7 +465,7 @@ static env_mode_flags_t compute_scope(set_cmd_opts_t &opts) {
 }
 
 /// Print the names of all environment variables in the scope. It will include the values unless the
-/// `set --list` flag was used.
+/// `set --names` flag was used.
 static int builtin_set_list(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, wchar_t **argv,
                             parser_t &parser, io_streams_t &streams) {
     UNUSED(cmd);
@@ -475,19 +474,30 @@ static int builtin_set_list(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
     UNUSED(parser);
 
     bool names_only = opts.list;
-    wcstring_list_t names = env_get_names(compute_scope(opts));
+    wcstring_list_t names = parser.vars().get_names(compute_scope(opts));
     sort(names.begin(), names.end());
 
-    for (size_t i = 0; i < names.size(); i++) {
-        const wcstring key = names.at(i);
+    for (const auto &key : names) {
         const wcstring e_key = escape_string(key, 0);
         streams.out.append(e_key);
 
         if (!names_only) {
-            auto var = env_get(key, compute_scope(opts));
-            if (!var.missing_or_empty()) {
+            wcstring val;
+            if (opts.shorten_ok && key == L"history") {
+                history_t *history =
+                    &history_t::history_with_name(history_session_id(parser.vars()));
+                for (size_t i = 1; i < history->size() && val.size() < 64; i++) {
+                    if (i > 1) val += L' ';
+                    val += expand_escape_string(history->item_at_index(i).str());
+                }
+            } else {
+                auto var = parser.vars().get(key, compute_scope(opts));
+                if (!var.missing_or_empty()) {
+                    val = expand_escape_variable(*var);
+                }
+            }
+            if (!val.empty()) {
                 bool shorten = false;
-                wcstring val = expand_escape_variable(*var);
                 if (opts.shorten_ok && val.length() > 64) {
                     shorten = true;
                     val.resize(60);
@@ -495,7 +505,7 @@ static int builtin_set_list(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
                 streams.out.append(L" ");
                 streams.out.append(val);
 
-                if (shorten) streams.out.push_back(ellipsis_char);
+                if (shorten) streams.out.push_back(get_ellipsis_char());
             }
         }
 
@@ -517,23 +527,23 @@ static int builtin_set_query(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
         assert(dest);
 
         std::vector<long> indexes;
-        int idx_count = parse_index(indexes, dest, scope, streams);
+        int idx_count = parse_index(indexes, dest, scope, streams, parser.vars());
         if (idx_count == -1) {
             free(dest);
-            builtin_print_help(parser, streams, cmd, streams.err);
+            builtin_print_error_trailer(parser, streams.err, cmd);
             return STATUS_CMD_ERROR;
         }
 
         if (idx_count) {
             wcstring_list_t result;
-            auto dest_str = env_get(dest, scope);
+            auto dest_str = parser.vars().get(dest, scope);
             if (dest_str) dest_str->to_list(result);
 
             for (auto idx : indexes) {
-                if (idx < 1 || (size_t)idx > result.size()) retval++;
+                if (idx < 1 || static_cast<size_t>(idx) > result.size()) retval++;
             }
         } else {
-            if (! env_get(arg, scope)) retval++;
+            if (!parser.vars().get(arg, scope)) retval++;
         }
 
         free(dest);
@@ -542,7 +552,8 @@ static int builtin_set_query(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
     return retval;
 }
 
-static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams) {
+static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams,
+                       const environment_t &vars) {
     const wchar_t *scope_name;
     switch (scope) {
         case ENV_LOCAL: {
@@ -563,7 +574,7 @@ static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams
         }
     }
 
-    const auto var = env_get(var_name, scope);
+    const auto var = vars.get(var_name, scope);
     if (!var) {
         streams.out.append_format(_(L"$%ls: not set in %ls scope\n"), var_name, scope_name);
         return;
@@ -573,14 +584,18 @@ static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams
     wcstring_list_t vals = var->as_list();
     streams.out.append_format(_(L"$%ls: set in %ls scope, %ls, with %d elements\n"), var_name,
                               scope_name, exportv, vals.size());
+
     for (size_t i = 0; i < vals.size(); i++) {
         if (vals.size() > 100) {
-            if (i == 50) streams.out.append(L"...\n");
+            if (i == 50) {
+                // try to print a mid-line ellipsis because we are eliding lines not words
+                streams.out.append(get_ellipsis_char() > 256 ? L"\u22EF" : get_ellipsis_str());
+                streams.out.push_back(L'\n');
+            }
             if (i >= 50 && i < vals.size() - 50) continue;
         }
         const wcstring value = vals[i];
-        const wcstring escaped_val =
-            escape_string(value, ESCAPE_NO_QUOTED, STRING_STYLE_SCRIPT);
+        const wcstring escaped_val = escape_string(value, ESCAPE_NO_QUOTED, STRING_STYLE_SCRIPT);
         streams.out.append_format(_(L"$%ls[%d]: length=%d value=|%ls|\n"), var_name, i + 1,
                                   value.size(), escaped_val.c_str());
     }
@@ -590,14 +605,15 @@ static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams
 static int builtin_set_show(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, wchar_t **argv,
                             parser_t &parser, io_streams_t &streams) {
     UNUSED(opts);
-
+    auto &vars = parser.vars();
     if (argc == 0) {  // show all vars
-        wcstring_list_t names = env_get_names(ENV_USER);
+        wcstring_list_t names = parser.vars().get_names(ENV_USER);
         sort(names.begin(), names.end());
-        for (auto it : names) {
-            show_scope(it.c_str(), ENV_LOCAL, streams);
-            show_scope(it.c_str(), ENV_GLOBAL, streams);
-            show_scope(it.c_str(), ENV_UNIVERSAL, streams);
+        for (const auto &name : names) {
+            if (name == L"history") continue;
+            show_scope(name.c_str(), ENV_LOCAL, streams, vars);
+            show_scope(name.c_str(), ENV_GLOBAL, streams, vars);
+            show_scope(name.c_str(), ENV_UNIVERSAL, streams, vars);
             streams.out.push_back(L'\n');
         }
     } else {
@@ -609,16 +625,16 @@ static int builtin_set_show(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
                 continue;
             }
 
-            if (wcschr(arg, L'[')) {
+            if (std::wcschr(arg, L'[')) {
                 streams.err.append_format(
                     _(L"%ls: `set --show` does not allow slices with the var names\n"), cmd);
-                builtin_print_help(parser, streams, cmd, streams.err);
+                builtin_print_error_trailer(parser, streams.err, cmd);
                 return STATUS_CMD_ERROR;
             }
 
-            show_scope(arg, ENV_LOCAL, streams);
-            show_scope(arg, ENV_GLOBAL, streams);
-            show_scope(arg, ENV_UNIVERSAL, streams);
+            show_scope(arg, ENV_LOCAL, streams, vars);
+            show_scope(arg, ENV_GLOBAL, streams, vars);
+            show_scope(arg, ENV_UNIVERSAL, streams, vars);
             streams.out.push_back(L'\n');
         }
     }
@@ -631,7 +647,7 @@ static int builtin_set_erase(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
                              parser_t &parser, io_streams_t &streams) {
     if (argc != 1) {
         streams.err.append_format(BUILTIN_ERR_ARG_COUNT2, cmd, L"--erase", 1, argc);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_CMD_ERROR;
     }
 
@@ -639,37 +655,44 @@ static int builtin_set_erase(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
     wchar_t *dest = argv[0];
 
     std::vector<long> indexes;
-    int idx_count = parse_index(indexes, dest, scope, streams);
+    int idx_count = parse_index(indexes, dest, scope, streams, parser.vars());
     if (idx_count == -1) {
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_CMD_ERROR;
     }
 
     int retval;
     if (!valid_var_name(dest)) {
         streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, dest);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
+    std::vector<event_t> evts;
     if (idx_count == 0) {  // unset the var
-        retval = env_remove(dest, scope);
-        // Temporarily swallowing ENV_NOT_FOUND errors to prevent
-        // breaking all tests that unset variables that aren't set.
+        retval = parser.vars().remove(dest, scope, &evts);
+        // When a non-existent-variable is unset, return ENV_NOT_FOUND as $status
+        // but do not emit any errors at the console as a compromise between user
+        // friendliness and correctness.
         if (retval != ENV_NOT_FOUND) {
             handle_env_return(retval, cmd, dest, streams);
         }
     } else {  // remove just the specified indexes of the var
-        const auto dest_var = env_get(dest, scope);
+        const auto dest_var = parser.vars().get(dest, scope);
         if (!dest_var) return STATUS_CMD_ERROR;
         wcstring_list_t result;
         dest_var->to_list(result);
         erase_values(result, indexes);
-        retval = env_set_reporting_errors(cmd, dest, scope, result, streams);
+        retval = env_set_reporting_errors(cmd, dest, scope, result, streams, parser.vars(), &evts);
+    }
+
+    // Fire any events.
+    for (const auto &evt : evts) {
+        event_fire(parser, evt);
     }
 
     if (retval != STATUS_CMD_OK) return retval;
-    return check_global_scope_exists(cmd, opts, dest, streams);
+    return check_global_scope_exists(cmd, opts, dest, streams, parser);
 }
 
 /// This handles the common case of setting the entire var to a set of values.
@@ -677,15 +700,13 @@ static int set_var_array(const wchar_t *cmd, set_cmd_opts_t &opts, const wchar_t
                          wcstring_list_t &new_values, int argc, wchar_t **argv, parser_t &parser,
                          io_streams_t &streams) {
     UNUSED(cmd);
-    UNUSED(parser);
     UNUSED(streams);
 
     if (opts.prepend || opts.append) {
         if (opts.prepend) {
             for (int i = 0; i < argc; i++) new_values.push_back(argv[i]);
         }
-
-        auto var_str = env_get(varname, ENV_DEFAULT);
+        auto var_str = parser.vars().get(varname, ENV_DEFAULT);
         wcstring_list_t var_array;
         if (var_str) var_str->to_list(var_array);
         new_values.insert(new_values.end(), var_array.begin(), var_array.end());
@@ -709,7 +730,7 @@ static int set_var_slices(const wchar_t *cmd, set_cmd_opts_t &opts, const wchar_
     if (opts.append || opts.prepend) {
         streams.err.append_format(
             L"%ls: Cannot use --append or --prepend when assigning to a slice", cmd);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -719,7 +740,7 @@ static int set_var_slices(const wchar_t *cmd, set_cmd_opts_t &opts, const wchar_
     }
 
     int scope = compute_scope(opts);  // calculate the variable scope based on the provided options
-    const auto var_str = env_get(varname, scope);
+    const auto var_str = parser.vars().get(varname, scope);
     if (var_str) var_str->to_list(new_values);
 
     // Slice indexes have been calculated, do the actual work.
@@ -740,7 +761,7 @@ static int builtin_set_set(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, w
                            parser_t &parser, io_streams_t &streams) {
     if (argc == 0) {
         streams.err.append_format(BUILTIN_ERR_MIN_ARG_COUNT1, cmd, 1);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -750,15 +771,15 @@ static int builtin_set_set(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, w
     argc--;
 
     std::vector<long> indexes;
-    int idx_count = parse_index(indexes, varname, scope, streams);
+    int idx_count = parse_index(indexes, varname, scope, streams, parser.vars());
     if (idx_count == -1) {
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
     if (!valid_var_name(varname)) {
         streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, varname);
-        builtin_print_help(parser, streams, cmd, streams.err);
+        builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
 
@@ -774,14 +795,21 @@ static int builtin_set_set(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, w
     }
     if (retval != STATUS_CMD_OK) return retval;
 
-    retval = env_set_reporting_errors(cmd, varname, scope, new_values, streams);
+    std::vector<event_t> evts;
+    retval =
+        env_set_reporting_errors(cmd, varname, scope, new_values, streams, parser.vars(), &evts);
+    // Fire any events.
+    for (const auto &evt : evts) {
+        event_fire(parser, evt);
+    }
+
     if (retval != STATUS_CMD_OK) return retval;
-    return check_global_scope_exists(cmd, opts, varname, streams);
+    return check_global_scope_exists(cmd, opts, varname, streams, parser);
 }
 
 /// The set builtin creates, updates, and erases (removes, deletes) variables.
 int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
-    const int incoming_exit_status = proc_get_last_status();
+    const int incoming_exit_status = parser.get_last_status();
     wchar_t *cmd = argv[0];
     int argc = builtin_count_args(argv);
     set_cmd_opts_t opts;
@@ -793,7 +821,7 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     argc -= optind;
 
     if (opts.print_help) {
-        builtin_print_help(parser, streams, cmd, streams.out);
+        builtin_print_help(parser, streams, cmd);
         return STATUS_CMD_OK;
     }
 
