@@ -5,16 +5,14 @@
 #ifdef _WIN32
 #define PCRE2_STATIC
 #endif
-#include <errno.h>
-#include <limits.h>
-#include <stdarg.h>
-#include <stddef.h>
-#include <stdint.h>
-#include <stdlib.h>
-#include <wchar.h>
-#include <wctype.h>
-
 #include <algorithm>
+#include <cerrno>
+#include <climits>
+#include <cstdarg>
+#include <cstddef>
+#include <cstdint>
+#include <cstdlib>
+#include <cwchar>
 #include <cwctype>
 #include <iterator>
 #include <memory>
@@ -26,6 +24,7 @@
 #include "builtin.h"
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
+#include "future_feature_flags.h"
 #include "io.h"
 #include "parse_util.h"
 #include "pcre2.h"
@@ -36,8 +35,6 @@
 
 class parser_t;
 
-#define STRING_ERR_MISSING _(L"%ls: Expected argument\n")
-
 // How many bytes we read() at once.
 // Bash uses 128 here, so we do too (see READ_CHUNK_SIZE).
 // This should be about the size of a line.
@@ -54,7 +51,7 @@ static void string_error(io_streams_t &streams, const wchar_t *fmt, ...) {
 static void string_unknown_option(parser_t &parser, io_streams_t &streams, const wchar_t *subcmd,
                                   const wchar_t *opt) {
     string_error(streams, BUILTIN_ERR_UNKNOWN, subcmd, opt);
-    builtin_print_help(parser, streams, L"string", streams.err);
+    builtin_print_error_trailer(parser, streams.err, L"string");
 }
 
 // We read from stdin if we are the second or later process in a pipeline.
@@ -63,7 +60,7 @@ static bool string_args_from_stdin(const io_streams_t &streams) {
 }
 
 static const wchar_t *string_get_arg_argv(int *argidx, const wchar_t *const *argv) {
-    return argv && argv[*argidx] ? argv[(*argidx)++] : NULL;
+    return argv && argv[*argidx] ? argv[(*argidx)++] : nullptr;
 }
 
 // A helper type for extracting arguments from either argv or stdin.
@@ -75,8 +72,8 @@ class arg_iterator_t {
     int argidx_;
     // If not using argv, a string to store bytes that have been read but not yet returned.
     std::string buffer_;
-    // If set, when reading from a stream, split on zeros instead of newlines.
-    const bool split0_;
+    // If set, when reading from a stream, split on newlines.
+    const bool split_;
     // Backing storage for the next() string.
     wcstring storage_;
     const io_streams_t &streams_;
@@ -85,10 +82,9 @@ class arg_iterator_t {
     /// not. On true, the string is stored in storage_.
     bool get_arg_stdin() {
         assert(string_args_from_stdin(streams_) && "should not be reading from stdin");
-        // Read in chunks from fd until buffer has a line (or zero if split0_ is set).
-        const char sep = split0_ ? '\0' : '\n';
+        // Read in chunks from fd until buffer has a line (or the end if split_ is unset).
         size_t pos;
-        while ((pos = buffer_.find(sep)) == std::string::npos) {
+        while (!split_ || (pos = buffer_.find('\n')) == std::string::npos) {
             char buf[STRING_CHUNK_SIZE];
             long n = read_blocked(streams_.stdin_fd, buf, STRING_CHUNK_SIZE);
             if (n == 0) {
@@ -118,18 +114,18 @@ class arg_iterator_t {
 
    public:
     arg_iterator_t(const wchar_t *const *argv, int argidx, const io_streams_t &streams,
-                   bool split0 = false)
-        : argv_(argv), argidx_(argidx), split0_(split0), streams_(streams) {}
+                   bool split = true)
+        : argv_(argv), argidx_(argidx), split_(split), streams_(streams) {}
 
     const wcstring *nextstr() {
         if (string_args_from_stdin(streams_)) {
-            return get_arg_stdin() ? &storage_ : NULL;
+            return get_arg_stdin() ? &storage_ : nullptr;
         }
         if (auto arg = string_get_arg_argv(&argidx_, argv_)) {
             storage_ = arg;
             return &storage_;
         } else {
-            return NULL;
+            return nullptr;
         }
     }
 };
@@ -157,6 +153,7 @@ typedef struct {  //!OCLINT(too many fields)
     bool start_valid = false;
     bool style_valid = false;
     bool no_empty_valid = false;
+    bool no_trim_newlines_valid = false;
 
     bool all = false;
     bool entire = false;
@@ -171,6 +168,7 @@ typedef struct {  //!OCLINT(too many fields)
     bool regex = false;
     bool right = false;
     bool no_empty = false;
+    bool no_trim_newlines = false;
 
     long count = 0;
     long length = 0;
@@ -178,8 +176,8 @@ typedef struct {  //!OCLINT(too many fields)
     long start = 0;
 
     const wchar_t *chars_to_trim = L" \f\n\r\t";
-    const wchar_t *arg1 = NULL;
-    const wchar_t *arg2 = NULL;
+    const wchar_t *arg1 = nullptr;
+    const wchar_t *arg2 = nullptr;
 
     escape_string_style_t escape_style = STRING_STYLE_SCRIPT;
 } options_t;
@@ -190,13 +188,13 @@ static int handle_flag_1(wchar_t **argv, parser_t &parser, io_streams_t &streams
     const wchar_t *cmd = argv[0];
 
     if (opts->style_valid) {
-        if (wcscmp(w.woptarg, L"script") == 0) {
+        if (std::wcscmp(w.woptarg, L"script") == 0) {
             opts->escape_style = STRING_STYLE_SCRIPT;
-        } else if (wcscmp(w.woptarg, L"url") == 0) {
+        } else if (std::wcscmp(w.woptarg, L"url") == 0) {
             opts->escape_style = STRING_STYLE_URL;
-        } else if (wcscmp(w.woptarg, L"var") == 0) {
+        } else if (std::wcscmp(w.woptarg, L"var") == 0) {
             opts->escape_style = STRING_STYLE_VAR;
-        } else if (wcscmp(w.woptarg, L"regex") == 0) {
+        } else if (std::wcscmp(w.woptarg, L"regex") == 0) {
             opts->escape_style = STRING_STYLE_REGEX;
         } else {
             string_error(streams, _(L"%ls: Invalid escape style '%ls'\n"), cmd, w.woptarg);
@@ -214,6 +212,9 @@ static int handle_flag_N(wchar_t **argv, parser_t &parser, io_streams_t &streams
     if (opts->no_newline_valid) {
         opts->no_newline = true;
         return STATUS_CMD_OK;
+    } else if (opts->no_trim_newlines_valid) {
+        opts->no_trim_newlines = true;
+        return STATUS_CMD_OK;
     }
     string_unknown_option(parser, streams, argv[0], argv[w.woptind - 1]);
     return STATUS_INVALID_ARGS;
@@ -408,23 +409,34 @@ static wcstring construct_short_opts(options_t *opts) {  //!OCLINT(high npath co
     if (opts->right_valid) short_opts.append(L"r");
     if (opts->start_valid) short_opts.append(L"s:");
     if (opts->no_empty_valid) short_opts.append(L"n");
+    if (opts->no_trim_newlines_valid) short_opts.append(L"N");
     return short_opts;
 }
 
 // Note that several long flags share the same short flag. That is okay. The caller is expected
 // to indicate that a max of one of the long flags sharing a short flag is valid.
-// Remember: adjust share/functions/string.fish when `string` options change
-static const struct woption long_options[] = {
-    {L"all", no_argument, NULL, 'a'},         {L"chars", required_argument, NULL, 'c'},
-    {L"count", required_argument, NULL, 'n'}, {L"entire", no_argument, NULL, 'e'},
-    {L"filter", no_argument, NULL, 'f'},      {L"ignore-case", no_argument, NULL, 'i'},
-    {L"index", no_argument, NULL, 'n'},       {L"invert", no_argument, NULL, 'v'},
-    {L"left", no_argument, NULL, 'l'},        {L"length", required_argument, NULL, 'l'},
-    {L"max", required_argument, NULL, 'm'},   {L"no-empty", no_argument, NULL, 'n'},
-    {L"no-newline", no_argument, NULL, 'N'},  {L"no-quoted", no_argument, NULL, 'n'},
-    {L"quiet", no_argument, NULL, 'q'},       {L"regex", no_argument, NULL, 'r'},
-    {L"right", no_argument, NULL, 'r'},       {L"start", required_argument, NULL, 's'},
-    {L"style", required_argument, NULL, 1},   {NULL, 0, NULL, 0}};
+// Remember: adjust share/completions/string.fish when `string` options change
+static const struct woption long_options[] = {{L"all", no_argument, nullptr, 'a'},
+                                              {L"chars", required_argument, nullptr, 'c'},
+                                              {L"count", required_argument, nullptr, 'n'},
+                                              {L"entire", no_argument, nullptr, 'e'},
+                                              {L"filter", no_argument, nullptr, 'f'},
+                                              {L"ignore-case", no_argument, nullptr, 'i'},
+                                              {L"index", no_argument, nullptr, 'n'},
+                                              {L"invert", no_argument, nullptr, 'v'},
+                                              {L"left", no_argument, nullptr, 'l'},
+                                              {L"length", required_argument, nullptr, 'l'},
+                                              {L"max", required_argument, nullptr, 'm'},
+                                              {L"no-empty", no_argument, nullptr, 'n'},
+                                              {L"no-newline", no_argument, nullptr, 'N'},
+                                              {L"no-quoted", no_argument, nullptr, 'n'},
+                                              {L"quiet", no_argument, nullptr, 'q'},
+                                              {L"regex", no_argument, nullptr, 'r'},
+                                              {L"right", no_argument, nullptr, 'r'},
+                                              {L"start", required_argument, nullptr, 's'},
+                                              {L"style", required_argument, nullptr, 1},
+                                              {L"no-trim-newlines", no_argument, nullptr, 'N'},
+                                              {nullptr, 0, nullptr, 0}};
 
 static const std::unordered_map<char, decltype(*handle_flag_N)> flag_to_function = {
     {'N', handle_flag_N}, {'a', handle_flag_a}, {'c', handle_flag_c}, {'e', handle_flag_e},
@@ -440,13 +452,15 @@ static int parse_opts(options_t *opts, int *optind, int n_req_args, int argc, wc
     const wchar_t *short_options = short_opts.c_str();
     int opt;
     wgetopter_t w;
-    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, NULL)) != -1) {
+    while ((opt = w.wgetopt_long(argc, argv, short_options, long_options, nullptr)) != -1) {
         auto fn = flag_to_function.find(opt);
         if (fn != flag_to_function.end()) {
             int retval = fn->second(argv, parser, streams, w, opts);
             if (retval != STATUS_CMD_OK) return retval;
         } else if (opt == ':') {
-            string_error(streams, STRING_ERR_MISSING, cmd);
+            streams.err.append(L"string ");  // clone of string_error
+            builtin_missing_argument(parser, streams, cmd, argv[w.woptind - 1],
+                                     false /* print_hints */);
             return STATUS_INVALID_ARGS;
         } else if (opt == '?') {
             string_unknown_option(parser, streams, cmd, argv[w.woptind - 1]);
@@ -461,15 +475,16 @@ static int parse_opts(options_t *opts, int *optind, int n_req_args, int argc, wc
     // If the caller requires one or two mandatory args deal with that here.
     if (n_req_args) {
         opts->arg1 = string_get_arg_argv(optind, argv);
-        if (!opts->arg1) {
-            string_error(streams, STRING_ERR_MISSING, cmd);
+        if (!opts->arg1 && n_req_args == 1) {
+            string_error(streams, BUILTIN_ERR_ARG_COUNT0, cmd);
             return STATUS_INVALID_ARGS;
         }
     }
     if (n_req_args > 1) {
         opts->arg2 = string_get_arg_argv(optind, argv);
         if (!opts->arg2) {
-            string_error(streams, STRING_ERR_MISSING, cmd);
+            string_error(streams, BUILTIN_ERR_MIN_ARG_COUNT1, cmd, n_req_args,
+                         !!opts->arg2 + !!opts->arg1);
             return STATUS_INVALID_ARGS;
         }
     }
@@ -602,7 +617,7 @@ class string_matcher_t {
 
    public:
     string_matcher_t(options_t opts_, io_streams_t &streams_)
-        : opts(std::move(opts_)), streams(streams_), total_matched(0) {}
+        : opts(opts_), streams(streams_), total_matched(0) {}
 
     virtual ~string_matcher_t() = default;
     virtual bool report_matches(const wcstring &arg) = 0;
@@ -618,9 +633,7 @@ class wildcard_matcher_t : public string_matcher_t {
                        io_streams_t &streams)
         : string_matcher_t(opts, streams), wcpattern(parse_util_unescape_wildcards(pattern)) {
         if (opts.ignore_case) {
-            for (size_t i = 0; i < wcpattern.length(); i++) {
-                wcpattern[i] = towlower(wcpattern[i]);
-            }
+            wcpattern = wcstolower(std::move(wcpattern));
         }
         if (opts.entire) {
             if (!wcpattern.empty()) {
@@ -641,11 +654,7 @@ class wildcard_matcher_t : public string_matcher_t {
         bool match;
 
         if (opts.ignore_case) {
-            wcstring s = arg;
-            for (size_t i = 0; i < s.length(); i++) {
-                s[i] = towlower(s[i]);
-            }
-            match = wildcard_match(s, wcpattern, false);
+            match = wildcard_match(wcstolower(arg), wcpattern, false);
         } else {
             match = wildcard_match(arg, wcpattern, false);
         }
@@ -667,7 +676,8 @@ class wildcard_matcher_t : public string_matcher_t {
 
 static wcstring pcre2_strerror(int err_code) {
     wchar_t buf[128];
-    pcre2_get_error_message(err_code, (PCRE2_UCHAR *)buf, sizeof(buf) / sizeof(wchar_t));
+    pcre2_get_error_message(err_code, reinterpret_cast<PCRE2_UCHAR *>(buf),
+                            sizeof(buf) / sizeof(wchar_t));
     return buf;
 }
 
@@ -677,7 +687,7 @@ struct compiled_regex_t {
 
     compiled_regex_t(const wchar_t *argv0, const wcstring &pattern, bool ignore_case,
                      io_streams_t &streams)
-        : code(0), match(0) {
+        : code(nullptr), match(nullptr) {
         // Disable some sequences that can lead to security problems.
         uint32_t options = PCRE2_NEVER_UTF;
 #if PCRE2_CODE_UNIT_WIDTH < 32
@@ -687,10 +697,10 @@ struct compiled_regex_t {
         int err_code = 0;
         PCRE2_SIZE err_offset = 0;
 
-        code =
-            pcre2_compile(PCRE2_SPTR(pattern.c_str()), pattern.length(),
-                          options | (ignore_case ? PCRE2_CASELESS : 0), &err_code, &err_offset, 0);
-        if (code == 0) {
+        code = pcre2_compile(PCRE2_SPTR(pattern.c_str()), pattern.length(),
+                             options | (ignore_case ? PCRE2_CASELESS : 0), &err_code, &err_offset,
+                             nullptr);
+        if (code == nullptr) {
             string_error(streams, _(L"%ls: Regular expression compile error: %ls\n"), argv0,
                          pcre2_strerror(err_code).c_str());
             string_error(streams, L"%ls: %ls\n", argv0, pattern.c_str());
@@ -698,15 +708,15 @@ struct compiled_regex_t {
             return;
         }
 
-        match = pcre2_match_data_create_from_pattern(code, 0);
+        match = pcre2_match_data_create_from_pattern(code, nullptr);
         assert(match);
     }
 
     ~compiled_regex_t() {
-        if (match != 0) {
+        if (match != nullptr) {
             pcre2_match_data_free(match);
         }
-        if (code != 0) {
+        if (code != nullptr) {
             pcre2_code_free(code);
         }
     }
@@ -741,7 +751,7 @@ class pcre2_matcher_t : public string_matcher_t {
             return 0;
         }
 
-        if (opts.entire) {
+        if (opts.entire && !opts.quiet) {
             streams.out.append(arg);
             streams.out.push_back(L'\n');
         }
@@ -753,11 +763,10 @@ class pcre2_matcher_t : public string_matcher_t {
 
             if (begin != PCRE2_UNSET && end != PCRE2_UNSET && !opts.quiet) {
                 if (opts.index) {
-                    streams.out.append_format(L"%lu %lu", (unsigned long)(begin + 1),
-                                              (unsigned long)(end - begin));
+                    streams.out.append_format(L"%lu %lu", (begin + 1), (end - begin));
                 } else if (end > begin) {
                     // May have end < begin if \K is used.
-                    streams.out.append(wcstring(&arg[begin], end - begin));
+                    streams.out.append(arg.substr(begin, end - begin));
                 }
                 streams.out.push_back(L'\n');
             }
@@ -778,56 +787,45 @@ class pcre2_matcher_t : public string_matcher_t {
     bool report_matches(const wcstring &arg) override {
         // A return value of true means all is well (even if no matches were found), false indicates
         // an unrecoverable error.
-        if (regex.code == 0) {
+        if (regex.code == nullptr) {
             // pcre2_compile() failed.
             return false;
         }
 
-        int matched = 0;
-
         // See pcre2demo.c for an explanation of this logic.
         PCRE2_SIZE arglen = arg.length();
-        int rc = report_match(
-            arg, pcre2_match(regex.code, PCRE2_SPTR(arg.c_str()), arglen, 0, 0, regex.match, 0));
-        if (rc < 0) {  // pcre2 match error.
-            return false;
-        } else if (rc == 0) {  // no match
-            return true;
-        }
-        matched++;
-        total_matched++;
+        int rc = report_match(arg, pcre2_match(regex.code, PCRE2_SPTR(arg.c_str()), arglen, 0, 0,
+                                               regex.match, nullptr));
 
-        if (opts.invert_match) {
+        if (rc < 0 /* pcre2 error */)
+            return false;
+        else if (rc == 0 /* no match */)
             return true;
-        }
+        else
+            total_matched++;
+
+        if (opts.invert_match) return true;
 
         // Report any additional matches.
-        PCRE2_SIZE *ovector = pcre2_get_ovector_pointer(regex.match);
-        while (opts.all || matched == 0) {
+        for (auto *ovector = pcre2_get_ovector_pointer(regex.match); opts.all; total_matched++) {
             uint32_t options = 0;
             PCRE2_SIZE offset = ovector[1];  // start at end of previous match
 
             if (ovector[0] == ovector[1]) {
-                if (ovector[0] == arglen) {
-                    break;
-                }
+                if (ovector[0] == arglen) break;
                 options = PCRE2_NOTEMPTY_ATSTART | PCRE2_ANCHORED;
             }
 
             rc = report_match(arg, pcre2_match(regex.code, PCRE2_SPTR(arg.c_str()), arglen, offset,
-                                               options, regex.match, 0));
-            if (rc < 0) {
+                                               options, regex.match, nullptr));
+
+            if (rc < 0 /* pcre2 error */)
                 return false;
-            }
-            if (rc == 0) {
-                if (options == 0) {  // all matches found
-                    break;
-                }
+            else if (rc == 0 /* no matches */) {
+                if (options == 0 /* all matches found now */) break;
                 ovector[1] = offset + 1;
                 continue;
             }
-            matched++;
-            total_matched++;
         }
         return true;
     }
@@ -881,7 +879,7 @@ class string_replacer_t {
 
    public:
     string_replacer_t(const wchar_t *argv0_, options_t opts_, io_streams_t &streams_)
-        : argv0(argv0_), opts(std::move(opts_)), total_replaced(0), streams(streams_) {}
+        : argv0(argv0_), opts(opts_), total_replaced(0), streams(streams_) {}
 
     virtual ~string_replacer_t() = default;
     int replace_count() const { return total_replaced; }
@@ -894,10 +892,10 @@ class literal_replacer_t : public string_replacer_t {
     size_t patlen;
 
    public:
-    literal_replacer_t(const wchar_t *argv0, const wcstring &pattern_, const wchar_t *replacement_,
+    literal_replacer_t(const wchar_t *argv0, wcstring pattern_, const wchar_t *replacement_,
                        const options_t &opts, io_streams_t &streams)
         : string_replacer_t(argv0, opts, streams),
-          pattern(pattern_),
+          pattern(std::move(pattern_)),
           replacement(replacement_),
           patlen(pattern.length()) {}
 
@@ -934,8 +932,13 @@ class regex_replacer_t : public string_replacer_t {
     regex_replacer_t(const wchar_t *argv0, const wcstring &pattern, const wcstring &replacement_,
                      const options_t &opts, io_streams_t &streams)
         : string_replacer_t(argv0, opts, streams),
-          regex(argv0, pattern, opts.ignore_case, streams),
-          replacement(interpret_escapes(replacement_)) {}
+          regex(argv0, pattern, opts.ignore_case, streams) {
+        if (feature_test(features_t::string_replace_backslash)) {
+            replacement = replacement_;
+        } else {
+            replacement = interpret_escapes(replacement_);
+        }
+    }
 
     bool replace_matches(const wcstring &arg) override;
 };
@@ -950,7 +953,7 @@ bool literal_replacer_t::replace_matches(const wcstring &arg) {
         replacement_occurred = true;
         result = arg;
     } else {
-        auto &cmp_func = opts.ignore_case ? wcsncasecmp : wcsncmp;
+        auto &cmp_func = opts.ignore_case ? wcsncasecmp : std::wcsncmp;
         const wchar_t *cur = arg.c_str();
         const wchar_t *end = cur + arg.size();
         while (cur < end) {
@@ -985,7 +988,7 @@ bool regex_replacer_t::replace_matches(const wcstring &arg) {
                        (opts.all ? PCRE2_SUBSTITUTE_GLOBAL : 0);
     size_t arglen = arg.length();
     PCRE2_SIZE bufsize = (arglen == 0) ? 16 : 2 * arglen;
-    wchar_t *output = (wchar_t *)malloc(sizeof(wchar_t) * bufsize);
+    wchar_t *output = static_cast<wchar_t *>(malloc(sizeof(wchar_t) * bufsize));
     int pcre2_rc;
     PCRE2_SIZE outlen = bufsize;
 
@@ -996,26 +999,27 @@ bool regex_replacer_t::replace_matches(const wcstring &arg) {
         pcre2_rc = pcre2_substitute(regex.code, PCRE2_SPTR(arg.c_str()), arglen,
                                     0,  // start offset
                                     options, regex.match,
-                                    0,  // match context
+                                    nullptr,  // match_data
                                     PCRE2_SPTR(replacement->c_str()), replacement->length(),
-                                    (PCRE2_UCHAR *)output, &outlen);
+                                    reinterpret_cast<PCRE2_UCHAR *>(output), &outlen);
 
         if (pcre2_rc != PCRE2_ERROR_NOMEMORY || bufsize >= outlen) {
             done = true;
         } else {
             bufsize = outlen;
-            wchar_t *new_output = (wchar_t *)realloc(output, sizeof(wchar_t) * bufsize);
+            wchar_t *new_output =
+                static_cast<wchar_t *>(realloc(output, sizeof(wchar_t) * bufsize));
             if (new_output) output = new_output;
         }
     }
 
-    wcstring outstr(output, outlen);
     bool rc = true;
     if (pcre2_rc < 0) {
         string_error(streams, _(L"%ls: Regular expression substitute error: %ls\n"), argv0,
                      pcre2_strerror(pcre2_rc).c_str());
         rc = false;
     } else {
+        wcstring outstr(output, outlen);
         bool replacement_occurred = pcre2_rc > 0;
         if (!opts.quiet && (!opts.filter || replacement_occurred)) {
             streams.out.append(outstr);
@@ -1073,7 +1077,7 @@ static int string_split_maybe0(parser_t &parser, io_streams_t &streams, int argc
 
     wcstring_list_t splits;
     size_t arg_count = 0;
-    arg_iterator_t aiter(argv, optind, streams, is_split0);
+    arg_iterator_t aiter(argv, optind, streams, !is_split0);
     while (const wcstring *arg = aiter.nextstr()) {
         if (opts.right) {
             split_about(arg->rbegin(), arg->rend(), sep.rbegin(), sep.rend(), &splits, opts.max,
@@ -1087,14 +1091,21 @@ static int string_split_maybe0(parser_t &parser, io_streams_t &streams, int argc
 
     // If we are from the right, split_about gave us reversed strings, in reversed order!
     if (opts.right) {
-        for (size_t j = 0; j < splits.size(); j++) {
-            std::reverse(splits[j].begin(), splits[j].end());
+        for (auto &split : splits) {
+            std::reverse(split.begin(), split.end());
         }
         std::reverse(splits.begin(), splits.end());
     }
 
     const size_t split_count = splits.size();
     if (!opts.quiet) {
+        if (is_split0 && !splits.empty()) {
+            // split0 ignores a trailing \0, so a\0b\0 is two elements.
+            // In contrast to split, where a\nb\n is three - "a", "b" and "".
+            //
+            // Remove the last element if it is empty.
+            if (splits.back().empty()) splits.pop_back();
+        }
         auto &buff = streams.out.buffer();
         for (const wcstring &split : splits) {
             buff.append(split, separation_type_t::explicitly);
@@ -1113,6 +1124,29 @@ static int string_split0(parser_t &parser, io_streams_t &streams, int argc, wcha
     return string_split_maybe0(parser, streams, argc, argv, true /* is_split0 */);
 }
 
+static int string_collect(parser_t &parser, io_streams_t &streams, int argc, wchar_t **argv) {
+    options_t opts;
+    opts.no_trim_newlines_valid = true;
+    int optind;
+    int retval = parse_opts(&opts, &optind, 0, argc, argv, parser, streams);
+    if (retval != STATUS_CMD_OK) return retval;
+
+    auto &buff = streams.out.buffer();
+    arg_iterator_t aiter(argv, optind, streams, /* don't split */ false);
+    while (const wcstring *arg = aiter.nextstr()) {
+        auto begin = arg->cbegin(), end = arg->cend();
+        if (!opts.no_trim_newlines) {
+            while (end > begin && *(end - 1) == L'\n') {
+                --end;
+            }
+        }
+
+        buff.append(begin, end, separation_type_t::explicitly);
+    }
+
+    return buff.size() > 0 ? STATUS_CMD_OK : STATUS_CMD_ERROR;
+}
+
 // Helper function to abstract the repeat logic from string_repeat
 // returns the to_repeat string, repeated count times.
 static wcstring wcsrepeat(const wcstring &to_repeat, size_t count) {
@@ -1151,7 +1185,8 @@ static int string_repeat(parser_t &parser, io_streams_t &streams, int argc, wcha
     arg_iterator_t aiter(argv, optind, streams);
     if (const wcstring *word = aiter.nextstr()) {
         const bool limit_repeat =
-            (opts.max > 0 && word->length() * opts.count > (size_t)opts.max) || !opts.count;
+            (opts.max > 0 && word->length() * opts.count > static_cast<size_t>(opts.max)) ||
+            !opts.count;
         const wcstring repeated =
             limit_repeat ? wcsrepeat_until(*word, opts.max) : wcsrepeat(*word, opts.count);
         is_empty = repeated.empty();
@@ -1178,7 +1213,7 @@ static int string_sub(parser_t &parser, io_streams_t &streams, int argc, wchar_t
     int nsub = 0;
     arg_iterator_t aiter(argv, optind, streams);
     while (const wcstring *s = aiter.nextstr()) {
-        typedef wcstring::size_type size_type;
+        using size_type = wcstring::size_type;
         size_type pos = 0;
         size_type count = wcstring::npos;
         if (opts.start > 0) {
@@ -1293,7 +1328,8 @@ string_subcommands[] = {
     {L"length", &string_length}, {L"match", &string_match},       {L"replace", &string_replace},
     {L"split", &string_split},   {L"split0", &string_split0},     {L"sub", &string_sub},
     {L"trim", &string_trim},     {L"lower", &string_lower},       {L"upper", &string_upper},
-    {L"repeat", &string_repeat}, {L"unescape", &string_unescape}, {NULL, NULL}};
+    {L"repeat", &string_repeat}, {L"unescape", &string_unescape}, {L"collect", &string_collect},
+    {nullptr, nullptr}};
 
 /// The string builtin, for manipulating strings.
 int builtin_string(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
@@ -1301,25 +1337,30 @@ int builtin_string(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     int argc = builtin_count_args(argv);
     if (argc <= 1) {
         streams.err.append_format(BUILTIN_ERR_MISSING_SUBCMD, cmd);
-        builtin_print_help(parser, streams, L"string", streams.err);
+        builtin_print_error_trailer(parser, streams.err, L"string");
         return STATUS_INVALID_ARGS;
     }
 
-    if (wcscmp(argv[1], L"-h") == 0 || wcscmp(argv[1], L"--help") == 0) {
-        builtin_print_help(parser, streams, L"string", streams.out);
+    if (std::wcscmp(argv[1], L"-h") == 0 || std::wcscmp(argv[1], L"--help") == 0) {
+        builtin_print_help(parser, streams, L"string");
         return STATUS_CMD_OK;
     }
 
     const string_subcommand *subcmd = &string_subcommands[0];
-    while (subcmd->name != 0 && wcscmp(subcmd->name, argv[1]) != 0) {
+    while (subcmd->name != nullptr && std::wcscmp(subcmd->name, argv[1]) != 0) {
         subcmd++;
     }
     if (!subcmd->handler) {
         streams.err.append_format(BUILTIN_ERR_INVALID_SUBCMD, cmd, argv[1]);
-        builtin_print_help(parser, streams, L"string", streams.err);
+        builtin_print_error_trailer(parser, streams.err, L"string");
         return STATUS_INVALID_ARGS;
     }
 
+    if (argc >= 3 && (std::wcscmp(argv[2], L"-h") == 0 || std::wcscmp(argv[2], L"--help") == 0)) {
+        wcstring string_dash_subcommand = wcstring(argv[0]) + L"-" + argv[1];
+        builtin_print_help(parser, streams, string_dash_subcommand.c_str());
+        return STATUS_CMD_OK;
+    }
     argc--;
     argv++;
     return subcmd->handler(parser, streams, argc, argv);
