@@ -5,33 +5,27 @@
 
 #include <errno.h>
 #include <limits.h>
-#include <pthread.h>
+// Needed for va_list et al.
 #include <stdarg.h>  // IWYU pragma: keep
-#include <stddef.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <termios.h>
-#include <wchar.h>
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>  // IWYU pragma: keep
 #endif
 
 #include <algorithm>
+#include <atomic>
 #include <functional>
-#include <iterator>
 #include <memory>
 #include <mutex>
-#include <sstream>
 #include <string>
-#include <tuple>
-#include <type_traits>
-#include <unordered_map>
 #include <vector>
 
 #include "fallback.h"  // IWYU pragma: keep
 #include "maybe.h"
-#include "signal.h"  // IWYU pragma: keep
+
+// PATH_MAX may not exist.
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
 
 // Define a symbol we can use elsewhere in our code to determine if we're being built on MS Windows
 // under Cygwin.
@@ -94,8 +88,6 @@ typedef std::vector<wcstring> wcstring_list_t;
 // on Mac OS X. See http://www.unicode.org/faq/private_use.html.
 #define ENCODE_DIRECT_BASE (wchar_t)0xF600
 #define ENCODE_DIRECT_END (ENCODE_DIRECT_BASE + 256)
-#define INPUT_COMMON_BASE (wchar_t)0xF700
-#define INPUT_COMMON_END (INPUT_COMMON_BASE + 64)
 
 // NAME_MAX is not defined on Solaris
 #if !defined(NAME_MAX)
@@ -146,24 +138,6 @@ enum {
 };
 typedef unsigned int escape_flags_t;
 
-// Directions.
-enum selection_direction_t {
-    // Visual directions.
-    direction_north,
-    direction_east,
-    direction_south,
-    direction_west,
-    direction_page_north,
-    direction_page_south,
-
-    // Logical directions.
-    direction_next,
-    direction_prev,
-
-    // Special value that means deselect.
-    direction_deselect
-};
-
 /// Issue a debug message with printf-style string formating and automatic line breaking. The string
 /// will begin with the string \c program_name, followed by a colon and a whitespace.
 ///
@@ -180,15 +154,14 @@ enum selection_direction_t {
 ///
 /// will print the string 'fish: Pi = 3.141', given that debug_level is 1 or higher, and that
 /// program_name is 'fish'.
-void __attribute__((noinline)) debug_impl(int level, const char *msg, ...)
-    __attribute__((format(printf, 2, 3)));
-void __attribute__((noinline)) debug_impl(int level, const wchar_t *msg, ...);
+[[gnu::noinline, gnu::format(printf, 2, 3)]] void debug_impl(int level, const char *msg, ...);
+[[gnu::noinline]] void debug_impl(int level, const wchar_t *msg, ...);
 
 /// The verbosity level of fish. If a call to debug has a severity level higher than \c debug_level,
 /// it will not be printed.
-extern int debug_level;
+extern std::atomic<int> debug_level;
 
-inline bool should_debug(int level) { return level <= debug_level; }
+inline bool should_debug(int level) { return level <= debug_level.load(std::memory_order_relaxed); }
 
 #define debug(level, ...)                                            \
     do {                                                             \
@@ -203,19 +176,22 @@ extern struct termios shell_modes;
 
 /// The character to use where the text has been truncated. Is an ellipsis on unicode system and a $
 /// on other systems.
-extern wchar_t ellipsis_char;
+wchar_t get_ellipsis_char();
+
 /// The character or string to use where text has been truncated (ellipsis if possible, otherwise
 /// ...)
-extern const wchar_t *ellipsis_str;
+const wchar_t *get_ellipsis_str();
 
 /// Character representing an omitted newline at the end of text.
-extern wchar_t omitted_newline_char;
+const wchar_t *get_omitted_newline_str();
+int get_omitted_newline_width();
 
 /// Character used for the silent mode of the read command
-extern wchar_t obfuscation_read_char;
+wchar_t get_obfuscation_read_char();
 
 /// How many stack frames to show when a debug() call is made.
-extern int debug_stack_frames;
+int get_debug_stack_frames();
+void set_debug_stack_frames(int);
 
 /// Profiling flag. True if commands should be profiled.
 extern bool g_profiling_active;
@@ -226,37 +202,18 @@ extern const wchar_t *program_name;
 /// Set to false if it's been determined we can't trust the last modified timestamp on the tty.
 extern const bool has_working_tty_timestamps;
 
-/// A list of all whitespace characters
-extern const wcstring whitespace;
-extern const char *whitespace_narrow;
-
-bool is_whitespace(const wchar_t input);
-bool is_whitespace(const wcstring &input);
-inline bool is_whitespace(const wchar_t *input) { return is_whitespace(wcstring(input)); }
-
-/// This macro is used to check that an argument is true. It is a bit like a non-fatal form of
-/// assert. Instead of exiting on failure, the current function is ended at once. The second
-/// parameter is the return value of the current function on failure.
-#define CHECK(arg, retval)                                                               \
-    if (!(arg)) {                                                                        \
-        debug(0, "function %s called with false value for argument %s", __func__, #arg); \
-        bugreport();                                                                     \
-        show_stackframe(L'E');                                                           \
-        return retval;                                                                   \
-    }
-
 // Pause for input, then exit the program. If supported, print a backtrace first.
 // The `return` will never be run  but silences oclint warnings. Especially when this is called
 // from within a `switch` block. As of the time I'm writing this oclint doesn't recognize the
 // `__attribute__((noreturn))` on the exit_without_destructors() function.
 // TODO: we use C++11 [[noreturn]] now, does that change things?
 #define FATAL_EXIT()                                \
-    {                                               \
+    do {                                            \
         char exit_read_buff;                        \
         show_stackframe(L'E');                      \
         ignore_result(read(0, &exit_read_buff, 1)); \
         exit_without_destructors(1);                \
-    }
+    } while (0)
 
 /// Exit the program at once after emitting an error message and stack trace if possible.
 /// We use our own private implementation of `assert()` for two reasons. First, some implementations
@@ -280,20 +237,8 @@ inline bool is_whitespace(const wchar_t *input) { return is_whitespace(wcstring(
 
 [[noreturn]] void __fish_assert(const char *msg, const char *file, size_t line, int error);
 
-/// Check if signals are blocked. If so, print an error message and return from the function
-/// performing this check.
-#define CHECK_BLOCK(retval)
-#if 0
-#define CHECK_BLOCK(retval)                                                \
-    if (signal_is_blocked()) {                                             \
-        debug(0, "function %s called while blocking signals. ", __func__); \
-        bugreport();                                                       \
-        show_stackframe(L'E');                                             \
-        return retval;                                                     \
-    }
-#endif
-
-/// Shorthand for wgettext call in situations where a C-style string is needed (e.g., fwprintf()).
+/// Shorthand for wgettext call in situations where a C-style string is needed (e.g.,
+/// std::fwprintf()).
 #define _(wstr) wgettext(wstr).c_str()
 
 /// Noop, used to tell xgettext that a string should be translated. Use this when a string cannot be
@@ -315,17 +260,19 @@ void vec_append(std::vector<T> &receiver, std::vector<T> &&donator) {
                     std::make_move_iterator(donator.end()));
 }
 
+/// Move an object into a shared_ptr.
+template <typename T>
+std::shared_ptr<T> move_to_sharedptr(T &&v) {
+    return std::make_shared<T>(std::move(v));
+}
+
+/// A function type to check for cancellation.
+/// \return true if execution should cancel.
+using cancel_checker_t = std::function<bool()>;
+
 /// Print a stack trace to stderr.
 void show_stackframe(const wchar_t msg_level, int frame_count = 100, int skip_levels = 0);
 
-/// Read a line from the stream f into the string. Returns the number of bytes read or -1 on
-/// failure.
-///
-/// If the carriage return character is encountered, it is ignored. fgetws() considers the line to
-/// end if reading the file results in either a newline (L'\n') character, the null (L'\\0')
-/// character or the end of file (WEOF) character.
-int fgetws2(wcstring *s, FILE *f);
-
 /// Returns a  wide character string equivalent of the specified multibyte character string.
 ///
 /// This function encodes illegal character sequences in a reversible way using the private use
@@ -344,121 +291,6 @@ char *wcs2str(const wchar_t *in);
 char *wcs2str(const wcstring &in);
 std::string wcs2string(const wcstring &input);
 
-/// Test if a string prefixes another. Returns true if a is a prefix of b.
-bool string_prefixes_string(const wcstring &proposed_prefix, const wcstring &value);
-bool string_prefixes_string(const wchar_t *proposed_prefix, const wcstring &value);
-bool string_prefixes_string(const wchar_t *proposed_prefix, const wchar_t *value);
-bool string_prefixes_string(const char *proposed_prefix, const std::string &value);
-bool string_prefixes_string(const char *proposed_prefix, const char *value);
-
-/// Test if a string is a suffix of another.
-bool string_suffixes_string(const wcstring &proposed_suffix, const wcstring &value);
-bool string_suffixes_string(const wchar_t *proposed_suffix, const wcstring &value);
-bool string_suffixes_string_case_insensitive(const wcstring &proposed_suffix,
-                                             const wcstring &value);
-
-/// Test if a string prefixes another without regard to case. Returns true if a is a prefix of b.
-bool string_prefixes_string_case_insensitive(const wcstring &proposed_prefix,
-                                             const wcstring &value);
-
-/// Case-insensitive string search, templated for use with both std::string and std::wstring.
-/// Modeled after std::string::find().
-/// \param fuzzy indicates this is being used for fuzzy matching and case insensitivity is
-/// expanded to include symbolic characters (#3584).
-/// \return the offset of the first case-insensitive matching instance of `needle` within
-/// `haystack`, or `string::npos()` if no results were found.
-template <typename T>
-size_t ifind(const T &haystack, const T &needle, bool fuzzy = false) {
-    using char_t = typename T::value_type;
-    auto locale = std::locale();
-
-    std::function<bool(char_t, char_t)> icase_eq;
-
-    if (!fuzzy) {
-        icase_eq = [&locale](char_t c1, char_t c2) {
-            return std::toupper(c1, locale) == std::toupper(c2, locale);
-        };
-    } else {
-        icase_eq = [&locale](char_t c1, char_t c2) {
-            // This `ifind()` call is being used for fuzzy string matching. Further extend case
-            // insensitivity to treat `-` and `_` as equal (#3584).
-
-            // The two lines below were tested to be 27% faster than
-            //      (c1 == '_' || c1 == '-') && (c2 == '-' || c2 == '_')
-            // while returning no false positives for all (c1, c2) combinations in the printable
-            // range (0x20-0x7E). It might return false positives outside that range, but fuzzy
-            // comparisons are typically called for file names only, which are unlikely to have
-            // such characters and this entire function is 100% broken on unicode so there's no
-            // point in worrying about anything outside of the ANSII range.
-            // ((c1 == Literal<char_t>('_') || c1 == Literal<char_t>('-')) &&
-            // ((c1 ^ c2) == (Literal<char_t>('-') ^ Literal<char_t>('_'))));
-
-            // One of the following would be an illegal comparison between a char and a wchar_t.
-            // However, placing them behind a constexpr gate results in the elision of the if
-            // statement and the incorrect branch, with the compiler's SFINAE support suppressing
-            // any errors in the branch not taken.
-            if (sizeof(char_t) == sizeof(char)) {
-                return std::toupper(c1, locale) == std::toupper(c2, locale) ||
-                ((c1 == '_' || c1 == '-') &&
-                ((c1 ^ c2) == ('-' ^ '_')));
-            } else {
-                return std::toupper(c1, locale) == std::toupper(c2, locale) ||
-                ((c1 == L'_' || c1 == L'-') &&
-                ((c1 ^ c2) == (L'-' ^ L'_')));
-            }
-        };
-    }
-
-    auto result =
-        std::search(haystack.begin(), haystack.end(), needle.begin(), needle.end(), icase_eq);
-    if (result != haystack.end()) {
-        return result - haystack.begin();
-    }
-    return T::npos;
-}
-
-/// Split a string by a separator character.
-wcstring_list_t split_string(const wcstring &val, wchar_t sep);
-
-/// Join a list of strings by a separator character.
-wcstring join_strings(const wcstring_list_t &vals, wchar_t sep);
-
-/// Support for iterating over a newline-separated string.
-template <typename Collection>
-class line_iterator_t {
-    // Storage for each line.
-    Collection storage;
-
-    // The collection we're iterating. Note we hold this by reference.
-    const Collection &coll;
-
-    // The current location in the iteration.
-    typename Collection::const_iterator current;
-
-public:
-    /// Construct from a collection (presumably std::string or std::wcstring).
-    line_iterator_t(const Collection &coll) : coll(coll), current(coll.cbegin()) {}
-
-    /// Access the storage in which the last line was stored.
-    const Collection &line() const {
-        return storage;
-    }
-
-    /// Advances to the next line. \return true on success, false if we have exhausted the string.
-    bool next() {
-        if (current == coll.end())
-            return false;
-        auto newline_or_end = std::find(current, coll.cend(), '\n');
-        storage.assign(current, newline_or_end);
-        current = newline_or_end;
-
-        // Skip the newline.
-        if (current != coll.cend())
-            ++current;
-        return true;
-    }
-};
-
 enum fuzzy_match_type_t {
     // We match the string exactly: FOOBAR matches FOOBAR.
     fuzzy_match_exact = 0,
@@ -492,7 +324,18 @@ static inline bool match_type_requires_full_replacement(fuzzy_match_type_t t) {
         case fuzzy_match_prefix: {
             return false;
         }
-        default: { return true; }
+        case fuzzy_match_case_insensitive:
+        case fuzzy_match_prefix_case_insensitive:
+        case fuzzy_match_substring:
+        case fuzzy_match_substring_case_insensitive:
+        case fuzzy_match_subsequence_insertions_only:
+        case fuzzy_match_none: {
+            return true;
+        }
+        default: {
+            DIE("Unreachable");
+            return false;
+        }
     }
 }
 
@@ -505,7 +348,16 @@ static inline bool match_type_shares_prefix(fuzzy_match_type_t t) {
         case fuzzy_match_prefix_case_insensitive: {
             return true;
         }
-        default: { return false; }
+        case fuzzy_match_substring:
+        case fuzzy_match_substring_case_insensitive:
+        case fuzzy_match_subsequence_insertions_only:
+        case fuzzy_match_none: {
+            return false;
+        }
+        default: {
+            DIE("Unreachabe");
+            return false;
+        }
     }
 }
 
@@ -543,39 +395,10 @@ void assert_is_background_thread(const char *who);
 #define ASSERT_IS_BACKGROUND_THREAD_TRAMPOLINE(x) assert_is_background_thread(x)
 #define ASSERT_IS_BACKGROUND_THREAD() ASSERT_IS_BACKGROUND_THREAD_TRAMPOLINE(__FUNCTION__)
 
-// fish_mutex is a wrapper around std::mutex that tracks whether it is locked, allowing for checking
-// if the mutex is locked. It owns a boolean guarded by the lock that records whether the lock is
-// currently locked; this is only used by assertions for correctness.
-class fish_mutex_t {
-    std::mutex lock_{};
-    bool is_locked_{false};
-
-   public:
-    constexpr fish_mutex_t() = default;
-    ~fish_mutex_t() = default;
-
-    void lock() {
-        lock_.lock();
-        is_locked_ = true;
-    }
-
-    void unlock() {
-        is_locked_ = false;
-        lock_.unlock();
-    }
-
-    // assert that this lock (identified as 'who') is locked in the function 'caller'.
-    void assert_is_locked(const char *who, const char *caller) const;
-
-    // return the underlying std::mutex. Note the fish_mutex_t cannot track locks to the underlying
-    // mutex; do not use assert_is_locked() with this.
-    std::mutex &get_mutex() { return lock_; }
-};
-
 /// Useful macro for asserting that a lock is locked. This doesn't check whether this thread locked
 /// it, which it would be nice if it did, but here it is anyways.
-void assert_is_locked(const fish_mutex_t &m, const char *who, const char *caller);
-#define ASSERT_IS_LOCKED(x) (x).assert_is_locked(#x, __FUNCTION__)
+void assert_is_locked(void *mutex, const char *who, const char *caller);
+#define ASSERT_IS_LOCKED(x) assert_is_locked((void *)(&x), #x, __FUNCTION__)
 
 /// Format the specified size (in bytes, kilobytes, etc.) into the specified stringbuffer.
 wcstring format_size(long long sz);
@@ -585,128 +408,23 @@ void format_size_safe(char buff[128], unsigned long long sz);
 
 /// Our crappier versions of debug which is guaranteed to not allocate any memory, or do anything
 /// other than call write(). This is useful after a call to fork() with threads.
-void debug_safe(int level, const char *msg, const char *param1 = NULL, const char *param2 = NULL,
-                const char *param3 = NULL, const char *param4 = NULL, const char *param5 = NULL,
-                const char *param6 = NULL, const char *param7 = NULL, const char *param8 = NULL,
-                const char *param9 = NULL, const char *param10 = NULL, const char *param11 = NULL,
-                const char *param12 = NULL);
+void debug_safe(int level, const char *msg, const char *param1 = nullptr,
+                const char *param2 = nullptr, const char *param3 = nullptr,
+                const char *param4 = nullptr, const char *param5 = nullptr,
+                const char *param6 = nullptr, const char *param7 = nullptr,
+                const char *param8 = nullptr, const char *param9 = nullptr,
+                const char *param10 = nullptr, const char *param11 = nullptr,
+                const char *param12 = nullptr);
 
 /// Writes out a long safely.
 void format_long_safe(char buff[64], long val);
 void format_long_safe(wchar_t buff[64], long val);
+void format_ullong_safe(wchar_t buff[64], unsigned long long val);
 
 /// "Narrows" a wide character string. This just grabs any ASCII characters and trunactes.
 void narrow_string_safe(char buff[64], const wchar_t *s);
 
-template <typename T>
-T from_string(const wcstring &x) {
-    T result;
-    std::wstringstream stream(x);
-    stream >> result;
-    return result;
-}
-
-template <typename T>
-T from_string(const std::string &x) {
-    T result = T();
-    std::stringstream stream(x);
-    stream >> result;
-    return result;
-}
-
-template <typename T>
-wcstring to_string(const T &x) {
-    std::wstringstream stream;
-    stream << x;
-    return stream.str();
-}
-
-// wstringstream is a huge memory pig. Let's provide some specializations where we can.
-template <>
-inline wcstring to_string(const long &x) {
-    wchar_t buff[128];
-    format_long_safe(buff, x);
-    return wcstring(buff);
-}
-
-template <>
-inline bool from_string(const std::string &x) {
-    return !x.empty() && strchr("YTyt1", x.at(0));
-}
-
-template <>
-inline bool from_string(const wcstring &x) {
-    return !x.empty() && wcschr(L"YTyt1", x.at(0));
-}
-
-template <>
-inline wcstring to_string(const int &x) {
-    return to_string(static_cast<long>(x));
-}
-
-wchar_t **make_null_terminated_array(const wcstring_list_t &lst);
-char **make_null_terminated_array(const std::vector<std::string> &lst);
-
-// Helper class for managing a null-terminated array of null-terminated strings (of some char type).
-template <typename CharType_t>
-class null_terminated_array_t {
-    CharType_t **array{NULL};
-
-    // No assignment or copying.
-    void operator=(null_terminated_array_t rhs);
-    null_terminated_array_t(const null_terminated_array_t &);
-
-    typedef std::vector<std::basic_string<CharType_t>> string_list_t;
-
-    size_t size() const {
-        size_t len = 0;
-        if (array != NULL) {
-            while (array[len] != NULL) {
-                len++;
-            }
-        }
-        return len;
-    }
-
-    void free(void) {
-        ::free((void *)array);
-        array = NULL;
-    }
-
-   public:
-    null_terminated_array_t() = default;
-
-    explicit null_terminated_array_t(const string_list_t &argv)
-        : array(make_null_terminated_array(argv)) {}
-
-    ~null_terminated_array_t() { this->free(); }
-
-    null_terminated_array_t(null_terminated_array_t &&rhs) : array(rhs.array) {
-        rhs.array = nullptr;
-    }
-
-    null_terminated_array_t operator=(null_terminated_array_t &&rhs) {
-        free();
-        array = rhs.array;
-        rhs.array = nullptr;
-    }
-
-    void set(const string_list_t &argv) {
-        this->free();
-        this->array = make_null_terminated_array(argv);
-    }
-
-    const CharType_t *const *get() const { return array; }
-    CharType_t **get() { return array; }
-
-    void clear() { this->free(); }
-};
-
-// Helper function to convert from a null_terminated_array_t<wchar_t> to a
-// null_terminated_array_t<char_t>.
-void convert_wide_array_to_narrow(const null_terminated_array_t<wchar_t> &arr,
-                                  null_terminated_array_t<char> *output);
-typedef std::lock_guard<fish_mutex_t> scoped_lock;
+typedef std::lock_guard<std::mutex> scoped_lock;
 typedef std::lock_guard<std::recursive_mutex> scoped_rlock;
 
 // An object wrapping a scoped lock and a value
@@ -719,15 +437,19 @@ typedef std::lock_guard<std::recursive_mutex> scoped_rlock;
 // Or for simple cases:
 //   name.acquire().value = "derp"
 //
-template <typename DATA>
+template <typename Data>
 class acquired_lock {
-    std::unique_lock<fish_mutex_t> lock;
-    acquired_lock(fish_mutex_t &lk, DATA *v) : lock(lk), value(v) {}
-
     template <typename T>
     friend class owning_lock;
 
-    DATA *value;
+    template <typename T>
+    friend class acquired_lock;
+
+    acquired_lock(std::mutex &lk, Data *v) : lock(lk), value(v) {}
+    acquired_lock(std::unique_lock<std::mutex> &&lk, Data *v) : lock(std::move(lk)), value(v) {}
+
+    std::unique_lock<std::mutex> lock;
+    Data *value;
 
    public:
     // No copying, move construction only
@@ -736,15 +458,30 @@ class acquired_lock {
     acquired_lock(acquired_lock &&) = default;
     acquired_lock &operator=(acquired_lock &&) = default;
 
-    DATA *operator->() { return value; }
-    const DATA *operator->() const { return value; }
-    DATA &operator*() { return *value; }
-    const DATA &operator*() const { return *value; }
+    Data *operator->() { return value; }
+    const Data *operator->() const { return value; }
+    Data &operator*() { return *value; }
+    const Data &operator*() const { return *value; }
+
+    /// Implicit conversion to const version.
+    operator acquired_lock<const Data>() {
+        // We're about to give up our lock, don't hold onto the data.
+        const Data *cvalue = value;
+        value = nullptr;
+        return acquired_lock<const Data>(std::move(lock), cvalue);
+    }
+
+    /// Create from a global lock.
+    /// This is used in weird cases where a global lock protects more than one piece of data.
+    static acquired_lock from_global(std::mutex &lk, Data *v) { return acquired_lock{lk, v}; }
+
+    /// \return a reference to the lock, for use with a condition variable.
+    std::unique_lock<std::mutex> &get_lock() { return lock; }
 };
 
 // A lock that owns a piece of data
 // Access to the data is only provided by taking the lock
-template <typename DATA>
+template <typename Data>
 class owning_lock {
     // No copying
     owning_lock &operator=(const scoped_lock &) = delete;
@@ -752,14 +489,15 @@ class owning_lock {
     owning_lock(owning_lock &&) = default;
     owning_lock &operator=(owning_lock &&) = default;
 
-    fish_mutex_t lock;
-    DATA data;
+    std::mutex lock;
+    Data data;
 
    public:
-    owning_lock(DATA &&d) : data(std::move(d)) {}
+    owning_lock(Data &&d) : data(std::move(d)) {}
+    owning_lock(const Data &d) : data(d) {}
     owning_lock() : data() {}
 
-    acquired_lock<DATA> acquire() { return {lock, &data}; }
+    acquired_lock<Data> acquire() { return {lock, &data}; }
 };
 
 /// A scoped manager to save the current value of some variable, and optionally set it to a new
@@ -815,6 +553,9 @@ class autoclose_fd_t {
         fd_ = fd;
     }
 
+    // \return if this has a valid fd.
+    bool valid() const { return fd_ >= 0; }
+
     autoclose_fd_t(const autoclose_fd_t &) = delete;
     void operator=(const autoclose_fd_t &) = delete;
     autoclose_fd_t(autoclose_fd_t &&rhs) : fd_(rhs.fd_) { rhs.fd_ = -1; }
@@ -828,13 +569,10 @@ class autoclose_fd_t {
     ~autoclose_fd_t() { close(); }
 };
 
-/// Appends a path component, with a / if necessary.
-void append_path_component(wcstring &path, const wcstring &component);
-
 wcstring format_string(const wchar_t *format, ...);
 wcstring vformat_string(const wchar_t *format, va_list va_orig);
 void append_format(wcstring &str, const wchar_t *format, ...);
-void append_formatv(wcstring &str, const wchar_t *format, va_list ap);
+void append_formatv(wcstring &target, const wchar_t *format, va_list va_orig);
 
 #ifdef HAVE_STD__MAKE_UNIQUE
 using std::make_unique;
@@ -850,20 +588,14 @@ std::unique_ptr<T> make_unique(Args &&... args) {
 /// character is detemrined by examining \c in. Returns 0 on error.
 ///
 /// \param in the position of the opening quote.
-wchar_t *quote_end(const wchar_t *in);
-
-/// A call to this function will reset the error counter. Some functions print out non-critical
-/// error messages. These should check the error_count before, and skip printing the message if
-/// MAX_ERROR_COUNT messages have been printed. The error_reset() should be called after each
-/// interactive command executes, to allow new messages to be printed.
-void error_reset();
+wchar_t *quote_end(const wchar_t *pos);
 
 /// This function should be called after calling `setlocale()` to perform fish specific locale
 /// initialization.
 void fish_setlocale();
 
-/// Call read while blocking the SIGCHLD signal. Should only be called if you _know_ there is data
-/// available for reading, or the program will hang until there is data.
+/// Call read, blocking and repeating on EINTR. Exits on EAGAIN.
+/// \return the number of bytes read, or 0 on EOF. On EAGAIN, returns -1 if nothing was read.
 long read_blocked(int fd, void *buf, size_t count);
 
 /// Loop a write request while failure is non-critical. Return -1 and set errno in case of critical
@@ -929,7 +661,7 @@ void common_handle_winch(int signal);
 /// Write the given paragraph of output, redoing linebreaks to fit the current screen.
 wcstring reformat_for_screen(const wcstring &msg);
 
-/// Make sure the specified direcotry exists. If needed, try to create it and any currently not
+/// Make sure the specified directory exists. If needed, try to create it and any currently not
 /// existing parent directories.
 ///
 /// \return 0 if, at the time of function return the directory exists, -1 otherwise.
@@ -963,18 +695,10 @@ void assert_is_not_forked_child(const char *who);
 #define ASSERT_IS_NOT_FORKED_CHILD_TRAMPOLINE(x) assert_is_not_forked_child(x)
 #define ASSERT_IS_NOT_FORKED_CHILD() ASSERT_IS_NOT_FORKED_CHILD_TRAMPOLINE(__FUNCTION__)
 
-/// Detect if we are Windows Subsystem for Linux by inspecting /proc/sys/kernel/osrelease
-/// and checking if "Microsoft" is in the first line.
+/// Determines if we are running under Microsoft's Windows Subsystem for Linux to work around
+/// some known limitations and/or bugs.
 /// See https://github.com/Microsoft/WSL/issues/423 and Microsoft/WSL#2997
-constexpr bool is_windows_subsystem_for_linux() {
-    // This function is called after fork() and before exec() in postfork.cpp. Make sure we
-    // don't allocate any memory here!
-#ifdef WSL
-    return true;
-#else
-    return false;
-#endif
-}
+bool is_windows_subsystem_for_linux();
 
 /// Detect if we are running under Cygwin or Cgywin64
 constexpr bool is_cygwin() {
@@ -986,7 +710,7 @@ constexpr bool is_cygwin() {
 }
 
 extern "C" {
-__attribute__((noinline)) void debug_thread_error(void);
+[[gnu::noinline]] void debug_thread_error(void);
 }
 
 /// Converts from wide char to digit in the specified base. If d is not a valid digit in the
@@ -1025,7 +749,7 @@ static T str_to_enum(const wchar_t *name, const enum_map<T> map[], int len) {
 
     while (left < right) {
         size_t mid = left + (right - left) / 2;
-        int cmp = wcscmp(name, map[mid].str);
+        int cmp = std::wcscmp(name, map[mid].str);
         if (cmp < 0) {
             right = mid;  // name was smaller than mid
         } else if (cmp > 0) {
@@ -1045,27 +769,13 @@ static const wchar_t *enum_to_str(T enum_val, const enum_map<T> map[]) {
             return entry->str;
         }
     }
-    return NULL;
+    return nullptr;
 };
 
-template <typename... Args>
-using tuple_list = std::vector<std::tuple<Args...>>;
-
-// Given a container mapping one X to many Y, return a list of {X,Y}
-template <typename X, typename Y>
-inline tuple_list<X, Y> flatten(const std::unordered_map<X, std::vector<Y>> &list) {
-    tuple_list<X, Y> results(list.size() * 1.5);  // just a guess as to the initial size
-    for (auto &kv : list) {
-        for (auto &v : kv.second) {
-            results.emplace_back(std::make_tuple(kv.first, v));
-        }
-    }
-
-    return results;
-}
-
 void redirect_tty_output();
 
+std::string get_path_to_tmp_dir();
+
 // Minimum allowed terminal size and default size if the detected size is not reasonable.
 #define MIN_TERM_COL 20
 #define MIN_TERM_ROW 2
@@ -1077,7 +787,6 @@ void invalidate_termsize(bool invalidate_vars = false);
 struct winsize get_current_winsize();
 
 bool valid_var_name_char(wchar_t chr);
-bool valid_var_name(const wchar_t *str);
 bool valid_var_name(const wcstring &str);
 bool valid_func_name(const wcstring &str);
 
@@ -1140,19 +849,19 @@ struct hash<const wcstring> {
 #endif
 
 /// Get the absolute path to the fish executable itself
-std::string get_executable_path(const char *fallback);
+std::string get_executable_path(const char *argv0);
 
 /// A RAII wrapper for resources that don't recur, so we don't have to create a separate RAII
 /// wrapper for each function. Avoids needing to call "return cleanup()" or similar / everywhere.
 struct cleanup_t {
-private:
+   private:
     const std::function<void()> cleanup;
-public:
-    cleanup_t(std::function<void()> exit_actions)
-        : cleanup{exit_actions} {}
-    ~cleanup_t() {
-        cleanup();
-    }
+
+   public:
+    cleanup_t(std::function<void()> exit_actions) : cleanup{std::move(exit_actions)} {}
+    ~cleanup_t() { cleanup(); }
 };
 
-#endif
+bool is_console_session();
+
+#endif  // FISH_COMMON_H
