@@ -113,8 +113,8 @@ static bool handler_matches(const event_handler_t &classv, const event_t &instan
             if (classv.desc.param1.pid == EVENT_ANY_PID) return true;
             return classv.desc.param1.pid == instance.desc.param1.pid;
         }
-        case event_type_t::job_exit: {
-            return classv.desc.param1.job_id == instance.desc.param1.job_id;
+        case event_type_t::caller_exit: {
+            return classv.desc.param1.caller_id == instance.desc.param1.caller_id;
         }
         case event_type_t::generic: {
             return classv.desc.str_param1 == instance.desc.str_param1;
@@ -167,13 +167,13 @@ wcstring event_get_desc(const event_t &evt) {
             DIE("Unreachable");
         }
 
-        case event_type_t::job_exit: {
-            job_t *j = job_t::from_job_id(ed.param1.job_id);
+        case event_type_t::caller_exit: {
+            job_t *j = job_t::from_job_id(ed.param1.caller_id);
             if (j) {
                 return format_string(_(L"exit handler for job %d, '%ls'"), j->job_id(),
                                      j->command_wcstr());
             } else {
-                return format_string(_(L"exit handler for job with job id %d"), ed.param1.job_id);
+                return format_string(_(L"exit handler for job with job id %d"), ed.param1.caller_id);
             }
             break;
         }
@@ -297,6 +297,8 @@ void event_fire_delayed(parser_t &parser) {
     auto &ld = parser.libdata();
     // Do not invoke new event handlers from within event handlers.
     if (ld.is_event) return;
+    // Do not invoke new event handlers if we are unwinding (#6649).
+    if (parser.get_cancel_signal()) return;
 
     std::vector<shared_ptr<event_t>> to_send;
     to_send.swap(ld.blocked_events);
@@ -351,7 +353,7 @@ struct event_type_name_t {
 static const event_type_name_t events_mapping[] = {{event_type_t::signal, L"signal"},
                                                    {event_type_t::variable, L"variable"},
                                                    {event_type_t::exit, L"exit"},
-                                                   {event_type_t::job_exit, L"job-id"},
+                                                   {event_type_t::caller_exit, L"job-id"},
                                                    {event_type_t::generic, L"generic"}};
 
 maybe_t<event_type_t> event_type_for_name(const wcstring &name) {
@@ -386,8 +388,8 @@ void event_print(io_streams_t &streams, maybe_t<event_type_t> type_filter) {
                           return d1.signal < d2.signal;
                       case event_type_t::exit:
                           return d1.param1.pid < d2.param1.pid;
-                      case event_type_t::job_exit:
-                          return d1.param1.job_id < d2.param1.job_id;
+                      case event_type_t::caller_exit:
+                          return d1.param1.caller_id < d2.param1.caller_id;
                       case event_type_t::variable:
                       case event_type_t::any:
                       case event_type_t::generic:
@@ -414,7 +416,7 @@ void event_print(io_streams_t &streams, maybe_t<event_type_t> type_filter) {
                                           evt->function_name.c_str());
                 break;
             case event_type_t::exit:
-            case event_type_t::job_exit:
+            case event_type_t::caller_exit:
                 streams.out.append_format(L"%d %ls\n", evt->desc.param1,
                                           evt->function_name.c_str());
                 break;
