@@ -20,41 +20,33 @@ struct function_properties_t {
     /// Parsed source containing the function.
     parsed_source_ref_t parsed_source;
 
-    /// Node containing the function body, pointing into parsed_source.
-    tnode_t<grammar::job_list> body_node;
+    /// Node containing the function statement, pointing into parsed_source.
+    /// We store block_statement, not job_list, so that comments attached to the header are
+    /// preserved.
+    tnode_t<grammar::block_statement> func_node;
 
     /// List of all named arguments for this function.
     wcstring_list_t named_arguments;
 
+    /// Mapping of all variables that were inherited from the function definition scope to their
+    /// values.
+    std::map<wcstring, wcstring_list_t> inherit_vars;
+
     /// Set to true if invoking this function shadows the variables of the underlying function.
-    bool shadow_scope;
+    bool shadow_scope{true};
 };
 
-/// Structure describing a function. This is used by the parser to store data on a function while
-/// parsing it. It is not used internally to store functions, the function_info_t structure
-/// is used for that purpose. Parhaps this should be merged with function_info_t.
-struct function_data_t {
-    /// Name of function.
-    wcstring name;
-    /// Description of function.
-    wcstring description;
-    /// List of all variables that are inherited from the function definition scope. The variable
-    /// values are snapshotted when function_add() is called.
-    wcstring_list_t inherit_vars;
-    /// Function's metadata fields
-    function_properties_t props;
-    /// List of all event handlers for this function.
-    std::vector<event_t> events;
-};
+using function_properties_ref_t = std::shared_ptr<const function_properties_t>;
 
 /// Add a function.
-void function_add(const function_data_t &data, const parser_t &parser);
+void function_add(wcstring name, wcstring description, function_properties_ref_t props,
+                  const wchar_t *filename);
 
 /// Remove the function with the specified name.
 void function_remove(const wcstring &name);
 
 /// Returns the properties for a function, or nullptr if none. This does not trigger autoloading.
-std::shared_ptr<const function_properties_t> function_get_properties(const wcstring &name);
+function_properties_ref_t function_get_properties(const wcstring &name);
 
 /// Returns by reference the definition of the function with the name \c name. Returns true if
 /// successful, false if no function with the given name exists.
@@ -67,17 +59,17 @@ bool function_get_definition(const wcstring &name, wcstring &out_definition);
 bool function_get_desc(const wcstring &name, wcstring &out_desc);
 
 /// Sets the description of the function with the name \c name.
-void function_set_desc(const wcstring &name, const wcstring &desc);
+void function_set_desc(const wcstring &name, const wcstring &desc, parser_t &parser);
 
 /// Returns true if the function with the name name exists.
 /// This may autoload.
-int function_exists(const wcstring &name);
+int function_exists(const wcstring &cmd, parser_t &parser);
 
 /// Attempts to load a function if not yet loaded. This is used by the completion machinery.
-void function_load(const wcstring &name);
+void function_load(const wcstring &cmd, parser_t &parser);
 
 /// Returns true if the function with the name name exists, without triggering autoload.
-int function_exists_no_autoload(const wcstring &name, const env_vars_snapshot_t &vars);
+int function_exists_no_autoload(const wcstring &cmd);
 
 /// Returns all function names.
 ///
@@ -100,18 +92,10 @@ const wchar_t *function_get_definition_file(const wcstring &name);
 /// This does not trigger autoloading.
 int function_get_definition_lineno(const wcstring &name);
 
-/// Returns a mapping of all variables of the specified function that were inherited from the scope
-/// of the function definition to their values.
-std::map<wcstring, env_var_t> function_get_inherit_vars(const wcstring &name);
-
 /// Creates a new function using the same definition as the specified function. Returns true if copy
 /// is successful.
 bool function_copy(const wcstring &name, const wcstring &new_name);
 
-/// Prepares the environment for executing a function.
-void function_prepare_environment(const wcstring &name, const wchar_t *const *argv,
-                                  const std::map<wcstring, env_var_t> &inherited_vars);
-
 /// Observes that fish_function_path has changed.
 void function_invalidate_path();
 
