@@ -117,6 +117,7 @@ static const input_function_metadata_t input_function_metadata[] = {
     {readline_cmd_t::history_token_search_backward, L"history-token-search-backward"},
     {readline_cmd_t::history_token_search_forward, L"history-token-search-forward"},
     {readline_cmd_t::self_insert, L"self-insert"},
+    {readline_cmd_t::self_insert_notfirst, L"self-insert-notfirst"},
     {readline_cmd_t::transpose_chars, L"transpose-chars"},
     {readline_cmd_t::transpose_words, L"transpose-words"},
     {readline_cmd_t::upcase_word, L"upcase-word"},
@@ -181,12 +182,13 @@ static wcstring input_get_bind_mode(const environment_t &vars) {
 }
 
 /// Set the current bind mode.
-static void input_set_bind_mode(env_stack_t &vars, const wcstring &bm) {
+static void input_set_bind_mode(parser_t &parser, const wcstring &bm) {
     // Only set this if it differs to not execute variable handlers all the time.
     // modes may not be empty - empty is a sentinel value meaning to not change the mode
     assert(!bm.empty());
-    if (input_get_bind_mode(vars) != bm) {
-        vars.set_one(FISH_BIND_MODE_VAR, ENV_GLOBAL, bm);
+    if (input_get_bind_mode(parser.vars()) != bm) {
+        // Must send events here - see #6653.
+        parser.set_var_and_fire(FISH_BIND_MODE_VAR, ENV_GLOBAL, bm);
     }
 }
 
@@ -360,7 +362,7 @@ void inputter_t::mapping_execute(const input_mapping_t &m, bool allow_commands)
 
     // !has_functions && !has_commands: only set bind mode
     if (!has_commands && !has_functions) {
-        if (!m.sets_mode.empty()) input_set_bind_mode(parser_->vars(), m.sets_mode);
+        if (!m.sets_mode.empty()) input_set_bind_mode(*parser_, m.sets_mode);
         return;
     }
 
@@ -399,7 +401,7 @@ void inputter_t::mapping_execute(const input_mapping_t &m, bool allow_commands)
     }
 
     // Empty bind mode indicates to not reset the mode (#2871)
-    if (!m.sets_mode.empty()) input_set_bind_mode(parser_->vars(), m.sets_mode);
+    if (!m.sets_mode.empty()) input_set_bind_mode(*parser_, m.sets_mode);
 }
 
 /// Try reading the specified function mapping.
@@ -496,7 +498,8 @@ char_event_t inputter_t::readch(bool allow_commands) {
 
         if (evt.is_readline()) {
             switch (evt.get_readline()) {
-                case readline_cmd_t::self_insert: {
+                case readline_cmd_t::self_insert:
+                case readline_cmd_t::self_insert_notfirst: {
                     // Typically self-insert is generated by the generic (empty) binding.
                     // However if it is generated by a real sequence, then insert that sequence.
                     for (auto iter = evt.seq.crbegin(); iter != evt.seq.crend(); ++iter) {
@@ -504,7 +507,13 @@ char_event_t inputter_t::readch(bool allow_commands) {
                     }
                     // Issue #1595: ensure we only insert characters, not readline functions. The
                     // common case is that this will be empty.
-                    return read_characters_no_readline();
+                    char_event_t res = read_characters_no_readline();
+
+                    // Hackish: mark the input style.
+                    res.input_style = evt.get_readline() == readline_cmd_t::self_insert_notfirst
+                                          ? char_input_style_t::notfirst
+                                          : char_input_style_t::normal;
+                    return res;
                 }
                 case readline_cmd_t::func_and: {
                     if (function_status_) {
