@@ -10,13 +10,10 @@ struct parse_token_t;
 
 namespace parse_productions {
 
-#define MAX_SYMBOLS_PER_PRODUCTION 6
-
 // A production is an array of unsigned char. Symbols are encoded directly as their symbol value.
 // Keywords are encoded with an offset of LAST_TOKEN_OR_SYMBOL + 1. So essentially we glom together
 // keywords and symbols.
 typedef uint8_t production_element_t;
-typedef production_element_t const production_t[MAX_SYMBOLS_PER_PRODUCTION];
 
 /// Resolve the type from a production element.
 inline parse_token_type_t production_element_type(production_element_t elem) {
@@ -44,8 +41,9 @@ inline bool production_element_is_valid(production_element_t elem) {
 
 /// Fetch a production. We are passed two input tokens. The first input token is guaranteed to not
 /// be invalid; the second token may be invalid if there's no more tokens. We may also set flags.
-const production_t *production_for_token(parse_token_type_t node_type, const parse_token_t &input1,
-                                         const parse_token_t &input2, uint8_t *out_tag);
-}
+const production_element_t *production_for_token(parse_token_type_t node_type,
+                                                 const parse_token_t &input1,
+                                                 const parse_token_t &input2, uint8_t *out_tag);
+}  // namespace parse_productions
 
 #endif
